@@ -64,12 +64,17 @@
 -define(ERAB_T_SESSION_DELETE, 2_000).
 
 -type teid() :: 0..16#ffffffff.
+-type addr() :: << _:32 >> | << _:128 >>.
+
+-type teid_addr() :: {teid(), %% GTP-U TEID
+                      addr()  %% GTP-U Transport Layer Address
+                     }.
 
 -record(erab_state, {from :: undefined | gen_statem:from(), %% destination to use when replying
-                     teid_from_access :: undefined | teid(), %% TEID to be used in eNB -> UPF
-                     teid_to_access :: undefined | teid(), %% TEID to be used in UPF -> eNB
-                     teid_from_core :: undefined | teid(), %% TEID to be used in PGW -> UPF
-                     teid_to_core :: undefined | teid(), %% TEID to be used in UPF -> PGW
+                     u2c :: undefined | teid_addr(), %% GTP-U params for UPF -> Core
+                     c2u :: undefined | teid_addr(), %% GTP-U params for UPF <- Core
+                     a2u :: undefined | teid_addr(), %% GTP-U params for UPF <- Access
+                     u2a :: undefined | teid_addr(), %% GTP-U params for UPF -> Access
                      seid_loc :: undefined | pfcp_peer:pfcp_seid(), %% local SEID, chosen by us
                      seid_rem :: undefined | pfcp_peer:pfcp_seid()  %% remote SEID, chosen by the UPF
                     }).
@@ -80,32 +85,35 @@
 %% public API
 %% ------------------------------------------------------------------
 
+%% @doc Allocate and start an E-RAB FSM process
+%% @param UID  *unique* E-RAB identifier.
+%% @returns process ID on success;  an error otherwise.
+%% @end
 -spec start_link(term()) -> gen_statem:start_ret().
 start_link(UID) ->
     gen_statem:start_link(?MODULE, [UID], []).
 
 
-%% @doc Indicate reception of E-RAB setup request (from the eNB).
+%% @doc Indicate reception of E-RAB setup request (from core).
 %% @param Pid  PID of an erab_fsm.
-%% @param UID  *unique* E-RAB identifier.
-%% @param TEID  TEID chosen by the eNB.
-%% @returns TEID to be sent to the core;  an error otherwise.
+%% @param F_TEID  TEID and bind address indicated by the MME.
+%% @returns TEID and Addr to be sent to the eNB;  an error otherwise.
 %% @end
--spec erab_setup_req(pid(), teid()) -> {ok, teid()} |
-                                       {error, term()}.
-erab_setup_req(Pid, TEID) ->
-    gen_statem:call(Pid, {?FUNCTION_NAME, TEID}).
+-spec erab_setup_req(pid(), teid_addr()) -> {ok, teid_addr()} |
+                                            {error, term()}.
+erab_setup_req(Pid, F_TEID) ->
+    gen_statem:call(Pid, {?FUNCTION_NAME, F_TEID}).
 
 
-%% @doc Indicate reception of E-RAB setup response (from the core).
+%% @doc Indicate reception of E-RAB setup response (from access).
 %% @param Pid  PID of an erab_fsm.
-%% @param TEID  TEID chosen by the core.
-%% @returns TEID to be sent to the eNB;  an error otherwise.
+%% @param F_TEID  TEID and bind address indicated by the eNB.
+%% @returns TEID and Addr to be sent to the MME;  an error otherwise.
 %% @end
--spec erab_setup_rsp(pid(), teid()) -> {ok, teid()} |
-                                       {error, term()}.
-erab_setup_rsp(Pid, TEID) ->
-    gen_statem:call(Pid, {?FUNCTION_NAME, TEID}).
+-spec erab_setup_rsp(pid(), teid_addr()) -> {ok, teid_addr()} |
+                                            {error, term()}.
+erab_setup_rsp(Pid, F_TEID) ->
+    gen_statem:call(Pid, {?FUNCTION_NAME, F_TEID}).
 
 
 -spec erab_release_req(pid()) -> ok.
@@ -139,18 +147,18 @@ callback_mode() ->
     [state_functions, state_enter].
 
 
-%% state WAIT_SETUP_REQ :: wait E-RAB SETUP Req from the Access
+%% state WAIT_SETUP_REQ :: wait E-RAB SETUP Req from core
 erab_wait_setup_req(enter, ?FUNCTION_NAME, S) ->
     ?LOG_DEBUG("State enter: ~p", [?FUNCTION_NAME]),
     {keep_state, S};
 
 erab_wait_setup_req({call, From},
-                    {erab_setup_req, TEID_FromAcc},
+                    {erab_setup_req, U2C},
                     #erab_state{} = S) ->
-    ?LOG_DEBUG("Rx E-RAB SETUP Req (Access TEID ~p)", [TEID_FromAcc]),
+    ?LOG_DEBUG("Rx E-RAB SETUP Req (U2C ~p)", [U2C]),
     {next_state, session_establish,
      S#erab_state{from = From,
-                  teid_from_access = TEID_FromAcc}};
+                  u2c = U2C}};
 
 erab_wait_setup_req(Event, EventData, S) ->
     ?LOG_ERROR("Unexpected event ~p: ~p", [Event, EventData]),
@@ -179,15 +187,16 @@ session_establish(info, #pfcp{} = PDU,
               seid = SEID_Rsp, %% matches F-SEID we sent in the ESTABLISH Req
               ie = #{pfcp_cause := 'Request accepted',
                      f_seid := #f_seid{seid = F_SEID},
-                     created_pdr := [#{f_teid := #f_teid{teid = TEID_ToCore}},
-                                     #{f_teid := #f_teid{teid = TEID_ToAcc}}]}} ->
-            ?LOG_DEBUG("PFCP session established"),
+                     created_pdr := [#{f_teid := F_TEID_C2U},
+                                     #{f_teid := F_TEID_A2U}]}} ->
+            C2U = {F_TEID_C2U#f_teid.teid, f_teid_addr(F_TEID_C2U)},
+            A2U = {F_TEID_A2U#f_teid.teid, f_teid_addr(F_TEID_A2U)},
+            ?LOG_DEBUG("PFCP session established (C2U ~p, A2U ~p)", [C2U, A2U]),
             {next_state, erab_wait_setup_rsp,
              S#erab_state{from = undefined,
                           seid_rem = F_SEID, %% SEID to be used in further requests from us
-                          teid_to_core = TEID_ToCore,
-                          teid_to_access = TEID_ToAcc},
-             [{reply, From, {ok, TEID_ToCore}}]};
+                          c2u = C2U, a2u = A2U},
+             [{reply, From, {ok, A2U}}]};
         _ ->
             ?LOG_ERROR("Rx unexpected PFCP PDU: ~p", [PDU]),
             {stop_and_reply,
@@ -200,7 +209,7 @@ session_establish(Event, EventData, S) ->
     {keep_state, S}.
 
 
-%% state WAIT_SETUP_RSP :: wait E-RAB SETUP Rsp from the Core
+%% state WAIT_SETUP_RSP :: wait E-RAB SETUP Rsp from access
 erab_wait_setup_rsp(enter, OldState, S) ->
     ?LOG_DEBUG("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
     {next_state, ?FUNCTION_NAME, S, %% loop transition to enable state_timeout
@@ -210,12 +219,12 @@ erab_wait_setup_rsp(state_timeout, _Timeout, _S) ->
     {stop, {shutdown, timeout}};
 
 erab_wait_setup_rsp({call, From},
-                    {erab_setup_rsp, TEID_FromCore},
+                    {erab_setup_rsp, U2A},
                     #erab_state{} = S) ->
-    ?LOG_DEBUG("Rx E-RAB SETUP Rsp (Core TEID ~p)", [TEID_FromCore]),
+    ?LOG_DEBUG("Rx E-RAB SETUP Rsp (U2A ~p)", [U2A]),
     {next_state, session_modify,
      S#erab_state{from = From,
-                  teid_from_core = TEID_FromCore}};
+                  u2a = U2A}};
 
 %% Catch-all handler for this state
 erab_wait_setup_rsp(Event, EventData, S) ->
@@ -240,7 +249,7 @@ session_modify(state_timeout, _Timeout,
 session_modify(info, #pfcp{} = PDU,
                #erab_state{from = From,
                            seid_loc = SEID_Rsp,
-                           teid_to_access = TEID_ToAcc} = S) ->
+                           c2u = C2U} = S) ->
     case PDU of
         #pfcp{type = session_modification_response,
               seid = SEID_Rsp, %% matches F-SEID we sent in the ESTABLISH Req
@@ -248,7 +257,7 @@ session_modify(info, #pfcp{} = PDU,
             ?LOG_DEBUG("PFCP session modified"),
             {next_state, erab_setup,
              S#erab_state{from = undefined},
-             [{reply, From, {ok, TEID_ToAcc}}]};
+             [{reply, From, {ok, C2U}}]};
         _ ->
             ?LOG_ERROR("Rx unexpected PFCP PDU: ~p", [PDU]),
             {stop_and_reply,
@@ -319,7 +328,7 @@ session_delete(Event, EventData, S) ->
     {keep_state, S}.
 
 
-%% state WAIT_RELEASE_RSP :: wait E-RAB RELEASE Rsp from the Core
+%% state WAIT_RELEASE_RSP :: wait E-RAB RELEASE Rsp
 erab_wait_release_rsp(enter, OldState, S) ->
     ?LOG_DEBUG("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
     {next_state, ?FUNCTION_NAME, S, %% loop transition to enable state_timeout
@@ -366,9 +375,20 @@ set_logging_prefix(UID) ->
     logger:set_process_metadata(#{prefix => Prefix}).
 
 
+-spec net_inst(atom()) -> undefined | binary().
+net_inst(Param) ->
+    case application:get_env(osmo_s1gw, Param) of
+        {ok, Val} -> Val;
+        undefined -> undefined
+    end.
+
+
 -spec session_establish_req(erab_state()) -> pfcp_peer:pfcp_session_rsp().
 session_establish_req(#erab_state{seid_loc = F_SEID, %% used as F-SEID
-                                  teid_from_access = TEID_FromAcc}) ->
+                                  u2c = U2C}) ->
+    %% Network Instance IE values for core/access
+    NI_Core = net_inst(pfcp_net_inst_core),
+    NI_Access = net_inst(pfcp_net_inst_access),
     %% Packet Detection Rules
     OHR = #outer_header_removal{header = 'GTP-U/UDP/IPv4'},
     PDRs = [#{pdr_id => {pdr_id, 1}, %% -- for Core -> Access
@@ -377,6 +397,7 @@ session_establish_req(#erab_state{seid_loc = F_SEID, %% used as F-SEID
               outer_header_removal => OHR,
               pdi => #{f_teid => #f_teid{teid = choose,
                                          ipv4 = choose},
+                       network_instance => NI_Core,
                        source_interface => {source_interface, 'Core'}}},
             #{pdr_id => {pdr_id, 2}, %% -- for Access -> Core
               far_id => {far_id, 2}, %% see FARs below
@@ -384,41 +405,57 @@ session_establish_req(#erab_state{seid_loc = F_SEID, %% used as F-SEID
               outer_header_removal => OHR,
               pdi => #{f_teid => #f_teid{teid = choose,
                                          ipv4 = choose},
+                       network_instance => NI_Access,
                        source_interface => {source_interface, 'Access'}}}],
     %% Forwarding Action Rules
-    OHC = #outer_header_creation{n6 = false,
-                                 n19 = false,
-                                 type = 'GTP-U',
-                                 teid = TEID_FromAcc,
-                                 ipv4 = <<127,0,0,1>>}, %% XXX: Core facing addr of the UPF
     FARs = [#{far_id => {far_id, 1}, %% -- for Core -> Access
-              %% We don't know the Core side TEID / GTP-U address yet, so we set
+              %% We don't know the Access side TEID / GTP-U address yet, so we set
               %% this FAR to DROP and modify it when we get E-RAB SETUP RESPONSE.
               apply_action => #{'DROP' => []}},
             #{far_id => {far_id, 2}, %% -- for Access -> Core
               apply_action => #{'FORW' => []},
               forwarding_parameters =>
-                #{outer_header_creation => OHC,
+                #{outer_header_creation => ohc(U2C),
+                  network_instance => NI_Core,
                   destination_interface => {destination_interface, 'Core'}}}],
     pfcp_peer:session_establish_req(F_SEID, PDRs, FARs).
 
 
 -spec session_modify_req(erab_state()) -> pfcp_peer:pfcp_session_rsp().
 session_modify_req(#erab_state{seid_rem = SEID, %% SEID allocated to us
-                               teid_from_core = TEID_FromCore}) ->
+                               u2a = U2A}) ->
+    %% Network Instance IE value for access
+    NI_Access = net_inst(pfcp_net_inst_access),
     %% Forwarding Action Rules
+    FARs = [#{far_id => {far_id, 1}, %% -- for Core -> Access
+              %% Now we know the Access side TEID / GTP-U address, so we modify
+              %% this FAR (which was previously set to DROP) to FORW.
+              apply_action => #{'FORW' => []},
+              update_forwarding_parameters =>
+                #{outer_header_creation => ohc(U2A),
+                  network_instance => NI_Access}}],
+    pfcp_peer:session_modify_req(SEID, [], FARs).
+
+
+ohc({TEID, Addr}) ->
     OHC = #outer_header_creation{n6 = false,
                                  n19 = false,
                                  type = 'GTP-U',
-                                 teid = TEID_FromCore,
-                                 ipv4 = <<127,0,0,1>>}, %% XXX: Access facing addr of the UPF
-    FARs = [#{far_id => {far_id, 1}, %% -- for Core -> Access
-              %% Now we know the Core side TEID / GTP-U address, so we modify
-              %% this FAR (which was previously set to DROP) to FORW.
-              apply_action => #{'FORW' => []},
-              forwarding_parameters =>
-                #{outer_header_creation => OHC,
-                  destination_interface => {destination_interface, 'Access'}}}],
-    pfcp_peer:session_modify_req(SEID, [], FARs).
+                                 teid = TEID},
+    case Addr of
+        << _:32 >> ->
+            OHC#outer_header_creation{ipv4 = Addr};
+        << _:128 >> ->
+            OHC#outer_header_creation{ipv6 = Addr}
+    end.
+
+
+%% select an address from F-TEID (prefer IPv6 over IPv4)
+f_teid_addr(#f_teid{ipv6 = << IPv6:16/bytes >>}) ->
+    IPv6;
+f_teid_addr(#f_teid{ipv4 = << IPv4:4/bytes >>}) ->
+    IPv4;
+f_teid_addr(#f_teid{}) ->
+    erlang:error("F-TEID IE contains no IPv4/IPv6").
 
 %% vim:set ts=4 sw=4 et:
